# Copyright 2024 PRIME team and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dapo import DAPORewardManager
from .naive import NaiveRewardManager
from .embodied import EmbodiedRewardManager
from .parallel import ParallelRewardManager
# Lazy import for embodied reward manager to avoid loading embodied dependencies for LLM/VLM tasks
def __getattr__(name):
    if name == "EmbodiedRewardManager":
        from .embodied import EmbodiedRewardManager
        return EmbodiedRewardManager
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

__all__ = ["DAPORewardManager", "NaiveRewardManager", "EmbodiedRewardManager","ParallelRewardManager"]
